/* Copyright 1997 Free Software Foundation, Inc.
   Contributed by Marcin Dalecki <dalecki@sub994.sub.uni-goettingen.de>

   This file is part of the Linux modutils.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>

#include "util.h"
#include "misc.h"
#include "conf_file.h"

struct SYMBOL *syms_hash[2048];	/* hash list of all visited symbols */
static struct MODULE *mods_first = NULL;	/* list of visited modules */

/*
 * Add the symbols exported by the kernel, by using a pseudo module called "-".
 * Return -1 if any error.
 */
static int
add_kernel_syms (void)
{
  int ret = -1;
  struct MODULE *mod;
  struct SYMBOL *tbsym[2000];
  int nbsym = 2;
  int requis;			/* Dummy */
  struct kernel_sym *ksym;
  int so_far = 0;
  int kernel_syms = 0;
  int size;

  /*
   * Create a pseudo module entry in the list of modules for the
   * information about the kernel exported symbols.
   */
  mod = (struct MODULE *) xmalloc (sizeof (struct MODULE));
  mod->next = mods_first;
  mods_first = mod;

  mod->name = xstrdup ("-");
  mod->pub.nb = mod->ext.nb = 0;
  mod->ext.tb = NULL;
  mod->pub.tb = NULL;

  /*
   * Fake the mod_use_count_ symbol which is provided by insmod
   */

  tbsym[0] = add_sym ("mod_use_count_", mod, SYM_DEFINED, &requis);
  tbsym[1] = add_sym ("__this_module", mod, SYM_DEFINED, &requis);

  load_kernel_symbols ();

  /*
   * Specification: depmod / kernel syms only
   * When initialising its symbol table from the kernel
   * depmod silently discards all symbol from loaded modules.
   *
   * This means that depmod may be used at any time to compute
   * the dependency table, even if there are modules already
   * loaded.
   *
   * depmod use the kernel system call to obtain the
   * symbol table, not /proc/ksyms. depmod assume that
   * kernel symbols are at the end of the list, just
   * after a pseudo symbol with a one character name: #
   */
  for (ksym = ksymtab; so_far < nksyms; ++so_far, ksym++)
    {
      if (kernel_syms)
	{
	  if (ksym->name[0] == '#')
	    continue;
	  /*
	   * FIXME: arbitrary limit!
	   */
	  assert (nbsym < 2000);
	  tbsym[nbsym++] = add_sym (ksym->name, mod, SYM_DEFINED, &requis);
	}
      else if (strcmp (ksym->name, "#") == 0)
	{
	  kernel_syms = 1;
	}
    }
  size = nbsym * sizeof (struct SYMBOL *);
  mod->pub.tb = (struct SYMBOL **) malloc (size);
  if (mod->pub.tb == NULL)
    lprintf ("can't allocate memory for kernel symbols");
  else
    {
      memcpy (mod->pub.tb, tbsym, size);
      ret = 0;
    }
  return ret;
}

/*
 * Format the dependency list of a module into a simple makefile
 */
static void
print_deps_file (const char *depfile, int verbose, int showerror)
{
  FILE *fout = stdout;
  struct MODULE *curr;

  if (depfile != NULL)
    {
      fout = fopen (depfile, "w");
      if (fout == NULL)
	{
	  lprintf ("can't open %s", depfile);
	  exit (1);
	}
    }

  /*
   * Loop through all modules we visited
   */
  for (curr = mods_first; curr; curr = curr->next)
    if (strcmp (curr->name, "-"))	/* ignore the kernel */
      {
	struct SYMBOL **ext = curr->ext.tb;
	int nbext = curr->ext.nb;
	int errs = 0;
	int e;

	struct dep_list
	  {
	    struct dep_list *next;
	    struct MODULE *mod;
	  }
	 *curr_deps = NULL;

	/*
	 * Go through the external symbols
	 */
	for (e = 0; e < nbext; e++, ext++)
	  {
	    struct MODULE *mod = (*ext)->module;

	    if (mod == NULL)
	      {
		/*
		 * No module associated with this external symbol, it's
		 * unresolved!!!!
		 */
		if (errs == 0)
		  lprintf ("*** Unresolved symbols in module %s", curr->name);
		if (showerror)
		  lprintf ("\t%s", (*ext)->name);
		errs++;
	      }
	    else if (strcmp (mod->name, "-") != 0)
	      {
		/*
		 * OK: We saw a module this symbol was defined in.
		 */
		struct dep_list *deps;

		/*
		 * Look if it's allready there
		 */
		for (deps = curr_deps; deps; deps = deps->next)
		  if (deps->mod == mod)
		    break;
		/*
		 * If not, add it to the current dependency list
		 */
		if (!deps)
		  {
		    deps = (struct dep_list *) xmalloc (sizeof (struct dep_list));
		    deps->next = curr_deps;
		    deps->mod = mod;
		    curr_deps = deps;
		  }
	      }
	  }
	if (errs == 0 && verbose)
	  printf ("%s\n", curr->name);
	fprintf (fout, "%s:", curr->name);

	/*
	 * Print the dependencies for this module.
	 */
	while (curr_deps)
	  {
	    struct dep_list *tmp;

	    fprintf (fout, " %s", curr_deps->mod->name);
	    tmp = curr_deps;
	    curr_deps = curr_deps->next;
	    free (tmp);
	  }
	/*
	 * The dependencies for modules are separated by empty lines.
	 */
	fprintf (fout, "\n\n");
      }
  if (fout != stdout)
    fclose (fout);
}


/*
   Lit les symboles d'un objet et enregistre dans syms
   Return -1 on error.
 */
static int
loadobj (const char *objname)
{
  struct MODULE *mod;

  mod = (struct MODULE *) xmalloc (sizeof (struct MODULE));
  mod->next = mods_first;
  mods_first = mod;

  mod->pub.nb = mod->ext.nb = 0;
  mod->ext.tb = NULL;
  mod->pub.tb = NULL;
  mod->name = xstrdup (objname);
  return load_obj (objname, mod);
}

/*
 * Print usage information ans exit.
 */
static void
usage (void)
{
  printf (
	   "Usage: depmod [-e -s -v ] -a [FORCED_KERNEL_VER]\n"
	   "       depmod [-e -s -v ] MODULE_1.o MODULE_2.o ...\n"
	   "Create module-dependency information for modprobe.\n\n"
	   "  -a, --all                  visit all modules\n"
	   "  -d, --debug                run in debug mode\n"
	   "  -e                         output unresolved symbols\n"
     "  -s, --system-log           use the system log for error reporting\n"
	   "      --help                 display this help and exit\n"
	   "  -v, --verbose              run in verbose mode\n"
	"  -V, --version              output version information and exit\n"
    );
}

int
main (int argc, char *argv[])
{
  int ret = -1;
  int showerror = 0;
  int verbose = 0;
  int stdmode = 0;

  int opt_tag;

  if (argc == 1)
    {
      usage ();
      return 1;
    }

  /*
   * List of symbols
   */
  memset (syms_hash, 0, sizeof (syms_hash));

  if (add_kernel_syms ())
    return 1;

  /*
   * Yes we are using getopts!
   */
  while (1)
    {
      static struct option long_opts[] =
      {
	{"all", 0, 0, 'a'},
	{"debug", 0, 0, 'd'},
	{"system-log", 0, 0, 's'},
	{"verbose", 0, 0, 'v'},
	{"version", 0, 0, 'V'},
	{"help", 0, 0, 'h'},
	{0, 0, 0, 0}		/* Table end tag */
      };
      int opt_ind = 0;

      opt_tag = getopt_long (argc, argv, "adesvV", long_opts,
			     &opt_ind);
      if (opt_tag == -1)
	break;

      switch (opt_tag)
	{
	case 'a':
	  stdmode = 1;
	  break;

	case 'd':
	  debugmode = 1;
	  break;

	case 'e':
	  showerror = 1;
	  break;

	case 's':
	  setsyslog ("depmod");	/* use the syslog for reporting */
	  break;

	case 'v':
	  verbose = 1;		/* give terse informations during run */
	  break;

	case '?':
	case 'h':
	  usage ();
	  exit (opt_tag == 'h' ? 1 : 0);
	  break;

	case 'V':
	  puts ("depmod (Linux modutils) " MODUTILS_VERSION);
	  if (argc != 2)
	    putchar ('\n');
	  break;

	default:
	  abort ();
	}
    }

  /*
   * Skip all automatically processed options.
   */
  argc -= optind;
  argv += optind;

  /*
   * argv now points to the first non-option argument
   * argc is the remaining argument count
   */

  if (stdmode)
    {
      struct mod_path *mods = NULL;
      struct mod_path *tmp;

      if (argc > 0)
	{
	  if (read_config_file (*argv))
	    {
	      lprintf ("%s does not exist", depfile);
	      return 1;
	    }
	}
      else if (read_config_file (NULL))		/* read the default config file */
	return 1;

      mods = find_matching_mods ("*", NULL, 1);
      ret = 0;
      while ((tmp = mods) && !ret)
	{
	  ret = loadobj (mods->path);
	  mods = mods->next;
	  free (tmp->path);
	  free (tmp);
	}
      while ((tmp = mods))
	{
	  mods = mods->next;
	  free (tmp->path);
	  free (tmp);
	}

      if (!ret)
	print_deps_file (depfile, verbose, showerror);
    }
  else
    {
      /*
       * Process all modules which are specified on the command line
       */

      for (ret = 0; argc > 0 && !ret; ++argv, --argc)
	ret = loadobj (*argv);
      if (!ret)
	print_deps_file (NULL, verbose, showerror);
    }

  return ret;
}
