/* Copyright 1996 Free Software Foundation, Inc.
   Contributed by Marcin Dalecki <dalecki@sub994.sub.uni-goettingen.de>

   This file is part of the Linux modutils.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <elf.h>
#include ELF_MACHINE_H

#include "util.h"
#include "misc.h"

#ifndef ElfW
#if ELFCLASSM == ELFCLASS32
#define ElfW(x)  Elf32_ ## x
#define ELFW(x)  ELF32_ ## x
#else
#define ElfW(x)  Elf64_ ## x
#define ELFW(x)  ELF64_ ## x
#endif
#endif

static void
setupmod (struct MODULE *mod, struct SYMBOL *tbpub[], int nbpub,
	  struct SYMBOL *tbext[], int nbext)
{
  mod->pub.nb = nbpub;
  mod->ext.nb = nbext;
  if (nbpub > 0)
    {
      int size = nbpub * sizeof (struct SYMBOL *);

      mod->pub.tb = (struct SYMBOL **) xmalloc (size);
      memcpy (mod->pub.tb, tbpub, size);
    }
  else
    {
      mod->pub.tb = NULL;
    }
  if (nbext > 0)
    {
      struct SYMBOL **ptext;
      int i;
      int size = nbext * sizeof (struct SYMBOL *);

      mod->ext.tb = (struct SYMBOL **) xmalloc (size);
      memcpy (mod->ext.tb, tbext, size);
      ptext = mod->ext.tb;
      for (i = 0; i < nbext; i++, ptext++)
	(*ptext)->req = 1;
    }
  else
    {
      mod->ext.tb = NULL;
    }
}

static char *
load_elf (FILE * fp, ElfW (Ehdr) * epnt, struct MODULE *mod)
{
  struct SYMBOL *tb_ext[5000];
  int nb_ext = 0;
  struct SYMBOL *tb_pub[5000];
  int nb_pub = 0;
  int mr = 0;
  ElfW (Shdr) * sections;
  ElfW (Shdr) * spnt;
  ElfW (Sym) * sp;
  ElfW (Sym) * symtab = NULL;
  long filesize;
  char *stringtab = (char *) 0;
  int i;
  int n;
  int nsymbols = 0;

  if (epnt->e_phnum != 0)
    return "can't handle program headers...";

  fseek (fp, 0L, SEEK_END);
  filesize = ftell (fp);

  /* load the section headers */
  sections = (ElfW (Shdr) *) xmalloc (epnt->e_shentsize * epnt->e_shnum);

  fseek (fp, epnt->e_shoff, SEEK_SET);
  fread ((char *) sections, epnt->e_shentsize * epnt->e_shnum, 1, fp);
  if (feof (fp) || ferror (fp))
    {
      free (sections);
      return "Error reading ELF section headers";
    }

  for (spnt = sections, i = 0; i < epnt->e_shnum; ++i, ++spnt)
    {
      switch (spnt->sh_type)
	{
	default:
	  break;		/* IGNORE */

	case SHT_SYMTAB:
	  if (nsymbols)
	    {
	      free (sections);
	      if (stringtab)
		free (stringtab);
	      return "can't handle >1 symbol section";
	    }
	  nsymbols = spnt->sh_size / spnt->sh_entsize;
	  symtab = (ElfW (Sym) *) xmalloc (nsymbols * sizeof (*symtab));
	  fseek (fp, spnt->sh_offset, SEEK_SET);
	  fread (symtab, nsymbols * sizeof (*symtab), 1, fp);
	  if (feof (fp) || ferror (fp))
	    {
	      free (sections);
	      free (symtab);
	      if (stringtab)
		free (stringtab);
	      return "Error reading ELF SYMTAB section";
	    }
	  break;

	case SHT_STRTAB:
	  stringtab = (char *) xmalloc (spnt->sh_size);
	  fseek (fp, spnt->sh_offset, SEEK_SET);
	  fread (stringtab, spnt->sh_size, 1, fp);
	  if (feof (fp) || ferror (fp))
	    {
	      free (sections);
	      free (stringtab);
	      if (symtab)
		free (symtab);
	      return "Error reading ELF STRTAB section";
	    }
	  break;
	}
    }

  for (n = nsymbols, sp = symtab; --n >= 0; sp++)
    {
      char *name = stringtab + sp->st_name;

      if (*name == '\0')
	continue;
      if (sp->st_shndx == SHN_UNDEF)
	tb_ext[nb_ext++] = add_sym (name, NULL, SYM_REQUIRED, &mr);
      else
	tb_pub[nb_pub++] = add_sym (name, mod, SYM_DEFINED, &mr);
    }

  setupmod (mod, tb_pub, nb_pub, tb_ext, nb_ext);
  if (stringtab)
    free (stringtab);
  if (symtab)
    free (symtab);
  free (sections);
  return (char *) 0;
}

int
load_obj (const char *module, struct MODULE *mod)
{
  FILE *fp;
  ElfW (Ehdr) header;
  char *errstr;

  /* open file and read header */
  if ((fp = fopen (module, "r")) == NULL)
    return 1;

  fread (&header, sizeof (ElfW (Ehdr)), 1, fp);
  if (feof (fp) || ferror (fp))
    return 1;

  errstr = load_elf (fp, &header, mod);

  if (errstr != (char *) 0)
    return 1;
  return 0;
}
