/* Copyright 1996 Free Software Foundation, Inc.
   Contributed by Marcin Dalecki <dalecki@sub994.sub.uni-goettingen.de>

   This file is part of the Linux modutils.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#include "util.h"
#include "misc.h"

/*
 * Strip the extension of a file name
 * Return a pointer to a static buffer
 */
const char *
strip_o (const char *fname)
{
  const char *pt = fname;

  while ((pt = strchr (fname, '/')))
    fname = pt + 1;
  if ((pt = strrchr (fname, '.'))
      && ((strcmp (pt, ".o") == 0) || (strcmp (pt, ".mod") == 0)))
    {
      char *leak = (char *) xmalloc (pt - fname + 1);
      strncpy (leak, fname, pt - fname);
      leak[pt - fname] = '\0';
      fname = leak;
    }
  return xstrdup (fname);
}

/*=================================================================*/

/*
 * Read and preprocess the contents of a configuration file.
 * Return the preprocessed contents on sccess and NULL otherwise.
 *
 * This functions implements efficient in place reading of files,
 * the usuall semantics of '#' comments and the arbitrary multiplicity of
 * white spaces as delimiters.
 */
char *
read_and_preprocess_file (const char *file)
{
  /*
   * If making any changes, please be shure to follow the exchaning roles
   * game between buf and fub strictly!
   */
  char *buf;
  char *fub;

  char *tmp;
  char *cp;
  FILE *fp;
  struct stat statb;

  if (stat (file, &statb) == -1
      || (statb.st_mode & S_IFMT) != S_IFREG
      || !(fp = fopen (file, "r")))
    {
      lprintf ("can't open file %s (%s)\n", file, strerror (errno));
      return NULL;
    }

  /*
   * OK: Now first read all of the file into an dynamically allocated buffer.
   */
  buf = (char *) xmalloc ((unsigned) (statb.st_size + 2));
  if (!fread (buf, sizeof (char), statb.st_size + 1, fp))
      lprintf ("Warning: file %s is empty!\n", file);
  buf[statb.st_size] = '\0';	/* be sure to NULL-terminate */
  fclose (fp);

  /*
   * Go through the whole buffer and remove comments by blanking them out.
   * Replace whitespaces other then '\n' by ' '.
   */
  for (cp = buf; *cp; ++cp)
    if (*cp == '#')
      while (*cp && *cp != '\n')
	{
	  *cp = ' ';
	  ++cp;
	}
    else if (isspace (*cp) && *cp != '\n')
      *cp = ' ';

  /*
   * Collapse all multiple white spaces into one
   */
  fub = (char *) xmalloc (strlen (buf) + 2);
  tmp = buf;
  for (cp = fub; *tmp; ++tmp)
    if (tmp[0] != ' ' || tmp[1] != ' ')
      {
	*cp = *tmp;
	++cp;
      }
  *cp = '\0';

  /*
   * Remove white spaces on line ends.
   */
  tmp = fub;
  for (cp = buf; *tmp; ++tmp)
    {
      if (tmp[0] == ' ' && tmp[1] == '\n')
	++tmp;
      *cp = *tmp;
      ++cp;
    }
  *cp = '\0';

  /*
   * Remove white space on line beginnings and the beginning of the file.
   */
  tmp = buf;
  if (*tmp == ' ')
    ++tmp;
  for (cp = fub; *tmp; ++tmp)
    {
      if (tmp[0] == '\n' && tmp[1] == ' ')
	{
	  ++tmp;
	  *tmp = '\n';
	}
      *cp = *tmp;
      ++cp;
    }
  *cp = '\0';

  /*
   * Make sure that every line is terminated by newline!
   */
  if ((cp > fub) && (*(cp - 1) != '\n'))
    {
      cp[0] = '\n';
      cp[1] = '\0';		/* no problem we allocated this bit */
    }

  free (buf);

  return fub;
}


/*
 * This function concatenates lines terminated by '\\' and increases "lines"
 * accordingly. If "src" there is no source or the source is empty, we are
 * returnning NULL. Otherwise we are returning the end of parsing.
 * the result is spaced in "src"! lines is adjusted to the number of lines
 * readed.
 */
char *
get_concat_line (char *src, int *lines)
{
  char *tmp;

  if (!src)			/* nothing to be prcessed */
    return NULL;

  tmp = src;

  if (!*tmp)
    return NULL;

  while (*src && (*src != '\n'))
    {
      if (src[0] == '\\' && src[1] == '\n')
	{
	  ++*lines;
	  ++src;
	  *src = '*';		/* fake it for the next loop */
	  --tmp;
	}
      else
	*tmp = *src;

      ++src;
      ++tmp;
    }
  *tmp = '\0';
  ++*lines;

  return src + 1;
}

/*=================================================================*/

/*
 * Interface to the symbol loading facilities
 */

int debugmode = 0;

struct kernel_sym *ksymtab = NULL;
int nksyms;

void
load_kernel_symbols (void)
{
  nksyms = get_kernel_syms (NULL);

  if (nksyms < 0)
    {
      lprintf ("get_kernel_sys failed: cannot find Kernel symbols!");
      exit (2);
    }

  if (ksymtab)
    free (ksymtab);

  if (nksyms)
    {
      ksymtab = (struct kernel_sym *) malloc (nksyms * sizeof *ksymtab);
      get_kernel_syms (ksymtab);
    }
}

/*=================================================================*/

/*
 * Error logging facilities.
 */

int log;			/* this is passed to insmod */
static int silent;

int errors;
const char *error_file;

void
error (const char *fmt,...)
{
  va_list args;

  if (silent)
    ;
  else if (log)
    {
      char buf[1024];
      int n;

      n = snprintf (buf, sizeof (buf), "%s: ",
		    error_file ? error_file : "insmod");
      va_start (args, fmt);
      vsnprintf (buf + n, sizeof (buf) - n, fmt, args);
      va_end (args);

      syslog (LOG_ERR, "%s", buf);
    }
  else
    {
      fprintf (stderr, "%s: ", error_file ? error_file : "insmod");
      va_start (args, fmt);
      vfprintf (stderr, fmt, args);
      va_end (args);
      putc ('\n', stderr);
    }

  errors++;
}

void
lprintf (const char *fmt,...)
{
  va_list args;

  if (silent)
    ;
  else if (log)
    {
      char buf[1024];
      va_start (args, fmt);
      vsnprintf (buf, sizeof (buf), fmt, args);
      va_end (args);
      syslog (LOG_INFO, "%s", buf);
    }
  else
    {
      va_start (args, fmt);
      vfprintf (stdout, fmt, args);
      va_end (args);
      putchar ('\n');
    }
}

void
setsyslog (const char *program)
{
  openlog (program, LOG_CONS, LOG_DAEMON);
  log = 1;
}
